#ifndef smart_ptr_HPP_HEADER_GUARD_
#define smart_ptr_HPP_HEADER_GUARD_
/*
For more details, which include example usage, full introduction, parameters, and tutorial,
see following link:
http://axter.com/smartptr
SET A SIDE NOTES:
For ease of testing and distribution, all the policies are included in the same header with the smart_ptr.
Once testing is near completion, the policies will be moved to separate header,  and the smart_ptr header 
will have an #include to the new policy headers.

This source file is documented in a Doxygen friendly format.
This format allows Doxygen to pull the comments out, and to create help
documents in HTML, RTF, man, and XML formats.
http://www.stack.nl/~dimitri/doxygen/index.html
*/

/*! 
@mainpage smart_ptr (Policy Based Smart Pointer)
@author David Maisonave (Axter) (609-345-1007) (<A HREF="http://www.axter.com" TARGET="_top">www.axter.com</A>)\n
Copyright (C) 2006\n
\n
@section Contents 
\ref DownloadSourceCode\n
\ref LastRevisionNumberDate\n
\ref SmartPtrLicense\n
\ref Why_Use_smart_ptr\n
\ref Policy_Classes\n
\ref Multithreading\n
\ref smart_ptr_Members\n 
\ref Motivation\n 

@section Introduction Introduction

smart_ptr is a smart pointer policy class that can use different 
ownership logic and semantic polices, which allows the developer to
get the best performance and/or interface for a particular requirement.
It has synchronization logic which allows it to be used in a \ref Multithreading
environment.

The smart_ptr class can be used with STL containers to create containers 
of smart pointers, moreover it can be used to create a container of 
abstract based objects via smart_ptr.
In general, smart_ptr is faster than boost::shared_ptr.  When used 
with STL containers, the smart pointer is faster than the boost 
pointer containers.  More importantly, the interface for an STL container 
of smart_ptr is a 100% compatible with STL containers, which is not the 
case with the boost pointer containers.\n
smart_ptr has a policy that allows it to synchronize access to both
the smart pointer and the pointee.  This allows both the smart_ptr and the pointee
to safely be used in a multithreading environment.\n
The smart_ptr has been compiled and tested on VC++ 6.0, 7.1, 8.0, GNU 3.x,
Borland 5.5, and Comeau 4.3.\n
\n
This help document includes three other smart pointers.\n
cow_ptr\n
copy_ptr\n
sync_ptr\n
These three smart pointers are not part of the smart_ptr class, but because
much of smart_ptr's implementation derived from these three classes, they're
included in the help document as a base of reference.  They're also used
in the unit testing as a base line comparison, and to insure that the 
smart_ptr policies are not effecting performance.\n
For more information about other types of smart pointers,
see \ref Other_Smart_Pointers
\n
@section  example_usage Example Usage
@include ../examples/example_smart_ptr_usage.cpp
\n
@section LastRevisionNumberDate Revision
\verbatim
$Date: 3/23/06 3:48a $ 
$Revision: 8 $
\endverbatim 
To view revison history, see \ref Revision_History
@section DownloadSourceCode Download
Download smart pointer source code using following link:\n
http://axter.com/smartptr/smartptr.zip
\n
Note:  All the smart pointers listed in this document are
implemented using *.hpp files.  There's no object files, 
*.cpp files, or DLL/SO files.  To use the smart pointers
in your project, just add an include.
\verbatim
#include "smart_ptr.hpp"
\endverbatim 
\n\n
Windows users can optionally use (CHM) help file, which
can be downloaded through the following link:\n
http://axter.com/smartptr/smartptr_chm.zip\n

*/


#ifndef __BORLANDC__ 
#if !defined(_MSC_VER) || (_MSC_VER > 1200)
//Both VC++6.0 and borland compilers don't support type defined via policy
#define SMART_PTR_SUPPORT_RETURN_TYPE_THROUGH_POLICY_
#endif
#endif

/*!
@struct allocator_default_policy
@brief This is the Default allocator policy for smart_ptr.
This policy does NOT require that the target class have a clone function or clone 
logic.
Cloning is performed by using the type pass to the constructor.
In order to use this policy, the target type must be pass to the constructor.
\n
<b>Example Usage</b>
@include ../examples/example_allocator_policy_usage.hxx
\n
@note
This policy can detect slicing better than the clone_function_allocator_policy,
but not as well as the clone_static_function_allocator_policy.
allocator_default_policy can detect slicing at runtime through constructor.
This policy is more generic than clone_function_allocator_policy and 
clone_static_function_allocator_policy, because it does not require the 
target class to have any type of clone function.
@see smart_ptr, clone_function_allocator_policy, 
clone_static_function_allocator_policy
*/
struct allocator_default_policy
{
/*!	
@brief allocate clones the input argument by calling the type copy constructor.
@pre Requires type to have a copy constructor, 
and pointee to be same type as pointer
@warning @image html warning.gif
Requires correct type for proper cloning
@param[in]  ptr valid pointer or NULL
@return A new object, or NULL if input parameter is NULL 
*/
	template<typename T_obj> static inline T_obj* allocate(const T_obj* ptr)
	{
		if (ptr) return new T_obj(*ptr);
		return NULL;
	}
	/*!
	@param[in]  ptr valid pointer or NULL
	*/
	template<typename T_obj> static inline void deallocate(T_obj* ptr){delete ptr;}
};

/*!	
@struct clone_function_allocator_policy
@brief This is an allocator policy that uses traditional cloning method.
This policy requires that the target class have a virtual clone function.
The normal usage is to create a base class with a pure virtual clone function, 
and have each derived class implement the clone function.  This includes 
derived derived classes, and so-on.
\n
<b>Example Usage</b>
@include ../examples/example_clone_function_allocator_policy_usage.hxx
\n
@note
This policy is least capable of detecting slicing.  Slicing can only be detected
on runtime when cloning is performed.  This policy also requires more 
maintenance, because each derived class has to implement the cloning function.
Because of this cloning function requirement, this policy is less generic than 
the allocator_default_policy
@see smart_ptr, allocator_default_policy, 
clone_static_function_allocator_policy
*/
struct clone_function_allocator_policy
{
/*!	
@brief allocate clones the input argument by calling virtual function do_clone.
@pre Base type requires do_clone virtual function, and pointee type requires 
do_clone implementation
@warning @image html warning.gif
slicing can occur if derived derived type fails to implement do_clone
@param[in]  ptr valid pointer or NULL
@return A new object, or NULL if input parameter is NULL 
*/
	template<typename T_obj> static inline T_obj* allocate(const T_obj* ptr)
	{
		if(!ptr) return NULL;
		return static_cast<T_obj*>(ptr->do_clone());
	}
	/*!
	@param[in]  ptr valid pointer or NULL
	*/
	template<typename T_obj> static inline void deallocate(T_obj* ptr){delete ptr;}
};

/*!	@struct clone_static_function_allocator_policy
@brief This is an allocator policy that uses a static cloning method.
This policy requires that the target class have a static clone function.
Each derived class must implement the static clone function.  
This includes derived derived classes, and so-on.
\n
<b>Example Usage</b>
@include ../examples/example_clone_static_function_allocator_policy_usage.hxx
\n
@note
This policy is able to avoid slicing better, because is can detect 
slicing at compile time.
This policy requires about the same amount of maintenance as the 
clone_function_allocator_policy, but more maintenance than the 
allocator_default_policy, because each derived class 
has to implement the static cloning function.
Because of this cloning function requirement, this policy is less generic than 
the allocator_default_policy
@see smart_ptr, allocator_default_policy, clone_function_allocator_policy, 
allocator_array_policy
*/
struct clone_static_function_allocator_policy
{
/*!	
@brief allocate clones the input argument by calling static function do_clone.
@pre Target type requires static do_clone function
@param[in]  ptr valid pointer or NULL
@return A new object, or NULL if input parameter is NULL 
*/
	template<typename T_obj> static inline T_obj* allocate(const T_obj* ptr)
	{
		if(!ptr) return NULL;
		return T_obj::do_clone(*ptr);
	}
	/*!
	@param[in]  ptr valid pointer or NULL
	*/
	template<typename T_obj> static inline void deallocate(T_obj* ptr){delete ptr;}
};

// @cond INCLUDE_ALL_OBJS_
/*!	@struct allocator_array_policy
@brief This is an allocator policy which creates an array of objects.
The current implementation of this policy is only practical in shared type 
ownership logic, or static size array. If a dynamic interchangeable array is 
needed, than a vector type should be used instead (smart_ptr<vector<T>, 
shared_ptr_policy<ref_link_policy> >
@warning @image html warning.gif
Due to the limitations of this policy, it may not make it through peer review
@see smart_ptr, allocator_default_policy, clone_function_allocator_policy, 
clone_static_function_allocator_policy
*/
template<size_t size_array>
struct allocator_array_policy
{
/*!	
@brief allocate clones an array.
@pre input argument must be pointing to an array equal to size_array
@warning @image html warning.gif
The allocate for this policy needs some work, to make it dynamic
@param[in]  ptr valid pointer or NULL
@return A new array of objects, or NULL if input parameter is NULL 
*/
	template<typename T_obj> static T_obj* allocate(const T_obj* ptr)
	{
		if (!ptr) return NULL;
#ifdef BOOST_ASSERT
		BOOST_ASSERT(size_array);
#endif //BOOST_ASSERT
		T_obj *arr = new T_obj[size_array];
		for(size_t i = 0;i < size_array;++i)
			arr[i] = ptr[i];
		return arr;
	}
	/*!
	@param[in]  ptr valid pointer or NULL
	*/
	template<typename T_obj> static inline void deallocate(T_obj* ptr){delete[] ptr;}
};
// @endcond 

/*!	@struct value_comparsion_semantic_policy
@brief This makes smart_ptr compare the pointee instead of the pointer address.
This is a policy that allows smart_ptr to be used with associated containers, 
and STL sorting algorithms.
When used with comparison operators, this policy makes 
smart_ptr compare the pointee, instead of the pointer address.
\n
<b>Example Usage</b>
@include ../examples/example_value_comparsion_semantic_policy_usage.hxx
\n
@note All members are static template functions, so as to avoid 
requiring the class to be a template class, and to avoid requiring
an instance of the class.
@see smart_ptr, pointer_comparsion_semantic_policy, no_comparsion_semantic_policy
*/
struct value_comparsion_semantic_policy
{
	template<class T1, class T2> static bool less_than(const T1 & a, const T2 & b){return (*a.c_ptr()) < (*b.c_ptr());}
	template<class T1, class T2> static bool greater_than(const T1 & a, const T2 & b){return (*a.c_ptr()) > (*b.c_ptr());}
	template<class T1, class T2> static bool less_than_or_equal(const T1 & a, const T2 & b){return (*a.c_ptr()) <= (*b.c_ptr());}
	template<class T1, class T2> static bool greater_than_or_equal(const T1 & a, const T2 & b){return (*a.c_ptr()) >= (*b.c_ptr());}
	template<class T1, class T2> static bool is_equal(const T1 & a, const T2 & b){return (*a.c_ptr()) == (*b.c_ptr());}
};

/*!	@struct pointer_comparsion_semantic_policy
@brief This makes smart_ptr compare the pointer address.
This policy gives the smart_ptr class normal pointer semantics when use with 
comparison operators. Comparison operators compare pointer address instead of 
pointee object. This may be desirable when using generic smart pointer coding, 
in which smart_ptr needs to work the same as other smart pointers.
@see smart_ptr, value_comparsion_semantic_policy, no_comparsion_semantic_policy
*/
struct pointer_comparsion_semantic_policy
{
	template<class T1, class T2> static bool less_than(const T1 & a, const T2 & b){return (a.c_ptr()) < (b.c_ptr());}
	template<class T1, class T2> static bool greater_than(const T1 & a, const T2 & b){return (a.c_ptr()) > (b.c_ptr());}
	template<class T1, class T2> static bool less_than_or_equal(const T1 & a, const T2 & b){return (a.c_ptr()) <= (b.c_ptr());}
	template<class T1, class T2> static bool greater_than_or_equal(const T1 & a, const T2 & b){return (a.c_ptr()) >= (b.c_ptr());}
	template<class T1, class T2> static bool is_equal(const T1 & a, const T2 & b){return (a.c_ptr()) == (b.c_ptr());}
};

/*!	@struct no_comparsion_semantic_policy
@brief This prevents comparison operators from being used with the smart_ptr.
Use this policy to disallow the use of comparison operators
@see smart_ptr, pointer_comparsion_semantic_policy, 
value_comparsion_semantic_policy
*/
struct no_comparsion_semantic_policy{};

/*!	@struct value_stream_operator_semantic_policy
@brief This is a pointee stream operator semantic policy.
Use this policy so as to allow stream operators to operate on the pointee
\n
<b>Example Usage</b>
@include ../examples/example_value_stream_operator_semantic_policy_usage.hxx
\n
@see smart_ptr, no_stream_operator_semantic_policy
*/
struct value_stream_operator_semantic_policy
{
	template<class T1, class T2> static T1& input_stream(T1 &is, T2 &smart_pointer){is >> (*smart_pointer.get_ptr());return is;}
	template<class T1, class T2> static T1& output_stream(T1 &io, T2 &smart_pointer){io << (*smart_pointer.c_ptr());return io;}
};

/*!	@struct no_stream_operator_semantic_policy
@brief This is a policy that prevents usage of stream operator.
Use this policy so as to disallow stream operator usage with the smart pointer
\n
\n
@see smart_ptr, value_stream_operator_semantic_policy
*/
struct no_stream_operator_semantic_policy{};

/*!	@struct value_arithmetic_semantic_policy
@brief This is a policy to allow arithmetic operators to target the pointee.
Use this policy to allow arithmetic operators to target the pointee 
instead of the pointer.  Normally, smart pointers try to emulate 
raw pointers, and when arithmetic operations are performed using 
the pointer, instead of what the pointer is pointing to.  This policy 
allows arithmetic operators to perform operations on the pointee, instead 
of the pointer.
\n
<b>Example Usage</b>
@include ../examples/example_no_arithmetic_semantic_policy_usage.hxx
\n
@see smart_ptr, no_arithmetic_semantic_policy
*/

struct value_arithmetic_semantic_policy
{
	// @cond INCLUDE_ALL_OBJS_
	template<class T> struct return_type{typedef T type;};
	// @endcond
	template<class T1, class T2> static T1& plus_equal_smart_ptr(T1 & a, const T2 & b){a.get_ptr()->operator+=(*b.c_ptr());return a;}
	template<class T1, class T2> static T1& plus_equal(T1 & a, const T2 & b){a.get_ptr()->operator+=(b);return a;}
	template<class T1, class T2> static T1& minus_equal_smart_ptr(T1 & a, const T2 & b){a.get_ptr()->operator-=(*b.c_ptr());return a;}
	template<class T1, class T2> static T1& minus_equal(T1 & a, const T2 & b){a.get_ptr()->operator-=(b);return a;}
	template<class T1, class T2> static T1 plus(const T1 & a, const T2 & b)
	{
		T1 c(a);
		c.get_ptr()->operator+=(*b.c_ptr());
		return c;
	}
	template<class T1, class T2> static T1 minus(const T1 & a, const T2 & b)
	{
		T1 c(a);
		c.get_ptr()->operator-=(*b.c_ptr());
		return c;
	}
};

// @cond INCLUDE_ALL_OBJS_
/*!	@struct no_arithmetic_semantic_policy
@brief A policy to prevents usage of arithmetic operators.
Use this policy to disallow the use of arithmetic operators
\n
@see smart_ptr, value_stream_operator_semantic_policy
\n
*/
struct no_arithmetic_semantic_policy{template<class T>struct return_type{typedef T type;};};
// @endcond

/*!
@struct ref_link_policy
@brief Use this policy to create a reference linking smart pointer.
In general, this policy is faster than the ref_count_policy,
but takes a little more memory.
It's not as easy to implement in synchronized code
compare to ref_count_policy.
It is more generic than ref_intrusive_policy, since it 
does not require that the target type have intrusive
functions.
Use ref_link_policy, when speed is the primary factor,
there is no synchronized referencing requirements, 
and there's no desire to implement intrusive functions
on the target type.
\n
<b>Example Usage</b>
@include ../examples/example_ref_link_policy_usage.hxx
\n
@note
The ref_link_policy can still be used in synchronized 
code, if intrusive_lock_policy is used.
@see smart_ptr, ref_count_policy, ref_intrusive_policy
*/
class ref_link_policy
{
	struct ref_link
	{
		ref_link():m_back(NULL), m_front(NULL){}
		ref_link *m_back;
		ref_link *m_front;
	private:  //Disallow copy constructor and assignment op
		ref_link(const ref_link&);
		ref_link& operator=(const ref_link&);
	};
	ref_link m_ref_link;
	void pop_link()
	{
		if (m_ref_link.m_back)
		{
			m_ref_link.m_back->m_front = m_ref_link.m_front;
		}
		if (m_ref_link.m_front)
		{
			m_ref_link.m_front->m_back = m_ref_link.m_back;
		}
		m_ref_link.m_front = NULL;
		m_ref_link.m_back = NULL;
	}
	void insert_link(ref_link *Target)
	{
		m_ref_link.m_back = Target;
		m_ref_link.m_front = Target->m_front;
		if (m_ref_link.m_front)
		{
			m_ref_link.m_front->m_back = &m_ref_link;
		}
		if (m_ref_link.m_back)
		{
			m_ref_link.m_back->m_front = &m_ref_link;
		}
	}
	ref_link_policy(const ref_link_policy&);
	ref_link_policy& operator=(const ref_link_policy&);
protected:
	template<class T>
		inline bool is_referenced(T&)const{return (m_ref_link.m_back || m_ref_link.m_front);}
public:
	ref_link_policy(){}
	template<class T1>
	inline ref_link_policy(ref_link_policy& src, T1&)
	{
		insert_link(&src.m_ref_link);
	}
	template<class T, class F, class LCK>
		void release(T& type, F clone_fct, LCK &lck_policy)
	{
		if (!is_referenced(type)){
			if (type && clone_fct)
			{
				clone_fct(type, false, NULL, NULL, NULL);
			}
			type = NULL;
		}
		else
		{
			pop_link();
			lck_policy.assignment_unlock_policy(type);
		}
	}
	template<class T1, class F1, class T2, class F2, class LCK>
		void assign(T1& type1, F1& clone_fct1, T2 type2, const F2 clone_fct2, ref_link_policy& src_policy, LCK &lck_policy)
	{
		if (!type1 || type1 != type2)
		{
			lck_policy.assignment_lock_policy(type1);
			const bool src_policy_eq_this = (this == &src_policy);
			release(type1, clone_fct1, lck_policy);
			type1 = type2;
			clone_fct1 = clone_fct2;
			if (src_policy_eq_this)
			{
				m_ref_link.m_back = NULL;
				m_ref_link.m_front = NULL;
			}
			else
				insert_link(&src_policy.m_ref_link);
		}
	}
};

/*!
@struct ref_count_policy
@brief Use this policy to create a reference counting smart pointer.
This policy is similar to reference counting logic used by 
boost::shared_ptr.
In general, it is slower than the other two reference 
policies (ref_link_policy and ref_intrusive_policy).
It is more generic than ref_intrusive_policy, since it 
does not require that the target type have intrusive
functions.
It takes less space than ref_link_policy, and it's 
easier to implement in synchronized code, than
ref_link_policy.
Use this policy when space is the primary factor, or
when referencing the same object in multiple threads 
and you don't wish to use an intrusive_lock_policy.
\n
<b>Example Usage</b>
@include ../examples/example_ref_count_policy_usage.hxx
\n
@see smart_ptr, ref_link_policy, ref_intrusive_policy
*/

class ref_count_policy
{
	struct ref_count
	{
		ref_count():m_count(new size_t(1)){}
		ref_count(size_t *count_):m_count(count_){}
		size_t *m_count;
	private:  //Disallow copy constructor and assignment op
		ref_count(const ref_count&);
		ref_count& operator=(const ref_count&);
	};
	ref_count m_ref_count;
	ref_count_policy(const ref_count_policy&);
	ref_count_policy& operator=(const ref_count_policy&);
protected:
	template<class T>
		inline bool is_referenced(T&)const{return (m_ref_count.m_count && (*m_ref_count.m_count) > 1);}
public:
	ref_count_policy(){}
	template<class T1>
		inline ref_count_policy(ref_count_policy& src, T1&):m_ref_count(src.m_ref_count.m_count)
	{
		if (m_ref_count.m_count) ++(*m_ref_count.m_count);
	}
	/// @todo Add assertion argument so function can call assertion policy
	template<class T, class F, class LCK>
		void release(T& type, F clone_fct, LCK &lck_policy)
	{
		if (!is_referenced(type)){
			if (type && clone_fct)
			{
				clone_fct(type, false, NULL, NULL, NULL);
			}
			type = NULL;
			delete m_ref_count.m_count;
			m_ref_count.m_count = NULL;
		}
		else
		{
			if (m_ref_count.m_count) --(*m_ref_count.m_count);
			lck_policy.assignment_unlock_policy(type);
		}
	}
	template<class T1, class F1, class T2, class F2, class LCK>
		void assign(T1& type1, F1& clone_fct1, T2 type2, const F2 clone_fct2, ref_count_policy& src_policy, LCK &lck_policy)
	{
		if (!type1 || type1 != type2)
		{
			lck_policy.assignment_lock_policy(type1);
			const bool src_policy_eq_this = (this == &src_policy);
			release(type1, clone_fct1, lck_policy);
			type1 = type2;
			clone_fct1 = clone_fct2;
			if (src_policy_eq_this)
				m_ref_count.m_count = new size_t(0); //Set to zero, because of increment below
			else
				m_ref_count.m_count = src_policy.m_ref_count.m_count;
			if (m_ref_count.m_count) ++(*m_ref_count.m_count);
		}
	}
};

/*!
@struct ref_intrusive_policy
@brief Use this policy to create an intrusive reference smart pointer.
In general, this policy is faster than both ref_count_policy 
and ref_link_policy.
However, it requires three intrusive functions that take
the target type pointer as a parameter.\n
void intrusive_ptr_add_ref(T *);\n
void intrusive_ptr_release(T *);\n
bool intrusive_ptr_is_ref(T *);\n
Example implementation for intrusive functions:\n
void intrusive_ptr_add_ref(Shape * p){++(p->references);}\n
void intrusive_ptr_release(Shape * p){\n
	if (--(p->references) == 0)\n
	delete p;\n
}\n
bool intrusive_ptr_is_ref(Shape * p){return (p->references > 1);}\n
\n
<b>Example Usage</b>
@include ../examples/example_ref_intrusive_policy_usage.hxx
\n
@note
The ref_link_policy can still be used in synchronized 
code, if intrusive_lock_policy is used.
@see smart_ptr, ref_count_policy, ref_link_policy
*/
class ref_intrusive_policy
{
	ref_intrusive_policy(const ref_intrusive_policy&);
	ref_intrusive_policy& operator=(const ref_intrusive_policy&);
protected:
	template<class T>
		inline bool is_referenced(T& type)const{return intrusive_ptr_is_ref(type);}
public:
	ref_intrusive_policy(){}
	template<class T>
	inline ref_intrusive_policy(ref_intrusive_policy&, T& type)
	{
		if (type) intrusive_ptr_add_ref(type);
	}
	template<class T, class F, class LCK>
		void release(T& type, F, LCK &)
	{
		//intrusive_ptr_release must perform assignment_unlock_policy (unlock)
		if (type) intrusive_ptr_release(type);
	}
	template<class T1, class F1, class T2, class F2, class LCK>
		void assign(T1& type1, F1& clone_fct1, T2 type2, const F2 clone_fct2, ref_intrusive_policy &src_policy, LCK &lck_policy)
	{
		if (!type1 || type1 != type2)
		{
			lck_policy.assignment_lock_policy(type1);
			const bool src_policy_eq_this = (this == &src_policy);
			release(type1, clone_fct1, lck_policy);
			type1 = type2;
			clone_fct1 = clone_fct2;
			if (type1 && !src_policy_eq_this) intrusive_ptr_add_ref(type1);
		}
	}
};

/*!
@struct no_lock_policy
@brief The no lock policy has no lock implementation.
Use this policy when synchronization is not required.
\n
@see smart_ptr, intrusive_lock_policy
*/
struct no_lock_policy
{
	no_lock_policy(){}
	template<class T_obj> inline void pointee_lock_policy(T_obj*){}
	template<class T_obj> inline void const_pointee_lock_policy(T_obj*){}
	template<class T_obj> inline void assignment_lock_policy(T_obj*){}
	template<class T_obj> inline void assignment_unlock_policy(T_obj*){}
	template<class T_obj> inline void destructor_lock_policy(T_obj*){}
	template<class T_obj> inline void destructor_unlock_policy(T_obj*){}
	template<class T_obj> inline T_obj* constructor_lock_policy(T_obj* p){return p;}
	template<class T_obj> inline void constructor_unlock_policy(T_obj*){}
	// @cond INCLUDE_ALL_OBJS_
	template<class T> struct return_type{typedef T* type_pointee;typedef T& type_ref;};//The struct type is here for a future optional feature
	// @endcond
	//A lock policy should have the following two functions implemented
	template<class T_obj> inline void lock(T_obj*){}
	template<class T_obj> inline void unlock(T_obj*){}
private:
	//A lock policy can optionally have the following two functions implemented and public
	template<class T_obj> inline bool trylock(T_obj*);
	template<class T_obj> inline bool islock(T_obj*);// Should only be used with trylock
	//The following should always be private with non implementation
	no_lock_policy(const no_lock_policy&);
	no_lock_policy& operator=(const no_lock_policy&);
};

/*!
@struct intrusive_lock_policy
@brief This implements an intrusive lock policy.
Allows synchronization of the smart pointer
and the pointee.\n
The intrusive_lock_policy requires the user
to implement two intrusive lock functions that
take the pointer type.\n
void intrusive_ptr_lock(T * p);\n
void intrusive_ptr_unlock(T * p);\n
Where T is the target type, and the functions 
are in a namespace accessible to the lock policy.\n
Optionally, the following functions can also 
be implemented, but are not required:\n
bool intrusive_ptr_trylock(T * p);
bool intrusive_ptr_islock(T * p);
\n
<b>Example Usage</b>
@include ../examples/example_intrusive_lock_policy_usage.hxx
\n
@see smart_ptr, no_lock_policy
\n
@note
The intrusive lock policy will work with ALL ownership policy types.
*/
struct intrusive_lock_policy
{
	intrusive_lock_policy(){}
	template<class T_obj> inline void pointee_lock_policy(T_obj* p){}
	template<class T_obj> inline void const_pointee_lock_policy(T_obj* p){}
	template<class T_obj> inline void assignment_lock_policy(T_obj* p){if (p) lock(p);}
	template<class T_obj> inline void assignment_unlock_policy(T_obj* p){if (p) unlock(p);}
	template<class T_obj> inline void destructor_lock_policy(T_obj* p){if (p) lock(p);}
	template<class T_obj> inline void destructor_unlock_policy(T_obj* p){if (p) unlock(p);}
	template<class T_obj> inline T_obj* constructor_lock_policy(T_obj* p){if (p) lock(p);return p;}
	template<class T_obj> inline void constructor_unlock_policy(T_obj* p){if (p) unlock(p);}
	template<class T_obj> inline void lock(T_obj* p){intrusive_ptr_lock(p);}
	template<class T_obj> inline void unlock(T_obj* p){intrusive_ptr_unlock(p);}
	template<class T_obj> inline bool trylock(T_obj* p){return intrusive_ptr_trylock(p);}
	template<class T_obj> inline bool islock(T_obj* p){return intrusive_ptr_islock(p);}// Should only be used with trylock
	// @cond INCLUDE_ALL_OBJS_
	template<class T> struct return_type{typedef T* type_pointee;typedef T& type_ref;};//The struct type is here for a future optional feature
	// @endcond
private: //The following should always be private with no implementation
	intrusive_lock_policy(const intrusive_lock_policy&);
	intrusive_lock_policy& operator=(const intrusive_lock_policy&);
};

//See Note#2 listed below
#ifndef SMART_PTR_EXCLUDE_DEFAULT_POLICY_SPECIFICATIONS_
typedef no_lock_policy												lock_default_policy;
typedef ref_link_policy												ref_default_policy;
#endif // not SMART_PTR_EXCLUDE_DEFAULT_POLICY_SPECIFICATIONS_

//weak_storage implementation is still in progress
struct weak_storage
{
	struct ref_count
	{
		ref_count():m_count(new size_t(1)){}
		size_t *m_count;
	private:  //Disallow copy constructor and assignment op
		ref_count(const ref_count&);
		ref_count& operator=(const ref_count&);
	};
	ref_count m_ref_count;
};

struct no_weak_support{};

/*!
@class copy_on_write_policy
@brief This is a copy on write ownership policy.
This policy is a mixture of a deep copy clone policy
and the shared pointer policy.
It shares the pointer as long as no attempt
is made to access the pointee through non-constant
operator->().\n
When access is made to the non-constant operator->(),
the smart pointer clones itself if it's being referenced.
When it clones itself, it performs a deep copy of the pointee.
\n
<b>Example Usage</b>
@include ../examples/example_copy_on_write_policy_usage.hxx
\n
@see smart_ptr, shared_ptr_policy, deep_copy_policy, ref_link_policy, 
ref_count_policy, ref_intrusive_policy, intrusive_lock_policy
\n
@note
The Ownership policy classes must take the lowest common 
denominator of arguments for constructor and common functions.  
Although this policy doesn't need the additional arguments, 
other policies do, and therefore to keep a standard interface, 
it must also take the additional arguments.
*/
template<class REFERENCE_POLICY = ref_default_policy, class LOCK_POLICY = lock_default_policy, class WEAK_SUPPORT_POLICY = no_weak_support>
class copy_on_write_policy : public REFERENCE_POLICY, public LOCK_POLICY
{
public:
	copy_on_write_policy(){}
	template<class T1, class F1, class T2>
		inline copy_on_write_policy(REFERENCE_POLICY& src, const T1& type, const F1&, const T2&):REFERENCE_POLICY(src, type)
	{
	}
	template<class T, class F>
		T get_non_constant_ptr(T& type, F clone_fct)
	{
		if (REFERENCE_POLICY::is_referenced(type) && type && clone_fct)
		{
			assign(type, clone_fct, 
				clone_fct(type, true, NULL, NULL, NULL), 
				clone_fct, *this, *this);
		}
		return type;
	}
};

/*!
@class shared_ptr_policy
@brief This is a shared pointer ownership policy.
This policy is similar to the ownership policy
used by boost::shared_ptr.  However, by default, 
the shared_ptr_policy uses reference link instead of
reference counting logic that is used by boost::shared_ptr.
In general, reference link logic is faster than
reference counting, so this make shared_ptr_policy with
it's default policies faster in general than
boost::shared_ptr.
For more information on pros and cons of the different
reference logic available, 
see ref_link_policy, ref_count_policy, and ref_intrusive_policy.
\n
<b>Example Usage</b>
@include ../examples/example_shared_ptr_policy_usage.hxx
\n
@see smart_ptr, copy_on_write_policy, deep_copy_policy, ref_link_policy, 
ref_count_policy, ref_intrusive_policy, intrusive_lock_policy
\n
@note
The Ownership policy classes must take the lowest common 
denominator of arguments for constructor and common functions.  
Although this policy doesn't need the additional arguments, 
other policies do, and therefore to keep a standard interface, 
it must also take the additional arguments.
Since shared_ptr_policy does not clone on write, get_non_constant_ptr
always returns the input pointer.
*/
template<class REFERENCE_POLICY = ref_default_policy, class LOCK_POLICY = lock_default_policy, class WEAK_SUPPORT_POLICY = no_weak_support>
class shared_ptr_policy : public REFERENCE_POLICY, public LOCK_POLICY
{
public:
	shared_ptr_policy(){}
	template<class T1, class F1, class T2>
		inline shared_ptr_policy(REFERENCE_POLICY& src, const T1& type, const F1&, const T2&):REFERENCE_POLICY(src, type)
	{
	}
	template<class T, class F>
		inline T& get_non_constant_ptr(T& type, const F&){return type;}
};

/*!
@class deep_copy_policy
@brief This is a deep copy (Clone) ownership policy.
This policy is similar to most clone smart pointers.
However, when use with default allocator policy, 
this type of policy does NOT require that the
pointee have a clone function implementation.
\n
<b>Example Usage</b>
@include ../examples/example_deep_copy_policy_usage.hxx
\n
@see smart_ptr, copy_on_write_policy, shared_ptr_policy, allocator_default_policy, 
clone_function_allocator_policy, clone_static_function_allocator_policy, 
intrusive_lock_policy
\n
*/
template<class LOCK_POLICY = lock_default_policy>
class deep_copy_policy : public LOCK_POLICY
{
public:
	deep_copy_policy(){}
	template<class T1, class F1, class T2>
		inline deep_copy_policy(deep_copy_policy&, T1 type_src, F1 clone_fct, T2& type_target)
	{
		type_target = clone_fct(type_src, true, NULL, NULL, NULL);
	}
	template<class T, class F, class LCK>
		inline void release(T& type, F clone_fct, LCK &)
	{
		if (type && clone_fct)
		{
			clone_fct(type, false, NULL, NULL, NULL);
		}
		type = NULL;
	}
	template<class T1, class F1, class T2, class F2, class LCK>
		void assign(T1& type1, F1& clone_fct1, T2 type2, const F2 clone_fct2, deep_copy_policy&, LCK &lck_policy)
	{
		if (!type1 && type1 != type2)
		{
			lck_policy.assignment_lock_policy(type1);
			release(type1, clone_fct1, lck_policy);
			if (type2)
				type1 = clone_fct2(type2, true, NULL, NULL, NULL);
			clone_fct1 = clone_fct2;
		}
	}
	template<class T, class F>
		inline T& get_non_constant_ptr(T& type, const F&)
	{//deep copy policy always clones on the constructor, so no need to clone here, and we can just return the pointer
		return type;
	}
};

/*! @struct no_checking_policy
@brief no_checking_policy has all checking functions stubbed out,
and performs no checking.
*/
struct no_checking_policy
{
	template<class T_obj> static void check_type_pass_to_constructor(T_obj*){}
	template<class T_obj> static void on_dereference(T_obj*){}
	template<class T_obj> static void on_const_dereference(T_obj*){}
	template<class T_obj> static void before_release(T_obj*){}
	template<class T_obj> static void after_release(T_obj*){}
};

// @cond INCLUDE_ALL_OBJS_
struct boost_assert_checking_policy
{
#ifdef BOOST_ASSERT
	template<class T_obj> static void check_type_pass_to_constructor(T_obj*){}
	template<class T_obj> static void on_dereference(T_obj*){}
	template<class T_obj> static void on_const_dereference(T_obj*){}
#else
	template<class T_obj> static void check_type_pass_to_constructor(T_obj* type)
	{
		BOOST_ASSERT(type != NULL);
		BOOST_ASSERT(typeid(*type) == typeid(T_obj));
	}
	template<class T_obj> static void on_dereference(T_obj* type)
	{
		BOOST_ASSERT(type != NULL);
		BOOST_ASSERT(typeid(*type) == typeid(T_obj));
	}
	template<class T_obj> static void on_const_dereference(T_obj* type)
	{
		BOOST_ASSERT(type != NULL);
		BOOST_ASSERT(typeid(*type) == typeid(T_obj));
	}
#endif //BOOST_ASSERT
	template<class T_obj> static void before_release(T_obj*){}
	template<class T_obj> static void after_release(T_obj*){}
};
// @endcond 


/*! @note
constructor_and_destructor_allocator_function is a function used
by the smart pointer that allows it to keep type information.
In a compliant C++ compiler, this function can be a static function
of the smart pointer class.  VC++ 6.0 is a prestandard compiler, 
and requires that this function be defined outside of the smart pointer
class.
*/
#if defined(_MSC_VER) && (_MSC_VER <= 1200)
template < typename TT, typename DT, typename AL> TT * constructor_and_destructor_allocator_function( TT *  ptr, bool bConstruct, DT*, AL* , void*) //Seen Note#1 below, for details on the last three arguments
{// The clone function method: (Thanks to Kai-Uwe Bux)
	if (!ptr) return NULL;
	if (bConstruct){return AL::allocate(static_cast<const DT*>(ptr));}
	AL::deallocate(ptr);
	return NULL;
}
#endif //defined(_MSC_VER) && (_MSC_VER <= 1200)


/*! @note
Note#2:\n
The following are the default policies for smart_ptr
These default policies can be replaced for a specific
translation unit, or for a specific project, by 
defining SMART_PTR_EXCLUDE_DEFAULT_POLICY_SPECIFICATIONS_
and by defining the desired policies.
To define custom default policies, do NOT use typedef.\n
Instead, use \#define.\n
If this entire package get separated into different *.hpp files, it 
than will be possible to use typedef instead of \#defines if the *.hpp
files are included in a certain order.
*/
#ifndef SMART_PTR_EXCLUDE_DEFAULT_POLICY_SPECIFICATIONS_
typedef value_comparsion_semantic_policy								comparison_default_policy;
typedef no_stream_operator_semantic_policy								stream_default_policy;
typedef no_checking_policy												checking_default_policy;
//The following two default policies are declared above the ownership polices
//typedef no_lock_policy												lock_default_policy;
//typedef ref_link_policy												ref_default_policy;
typedef copy_on_write_policy<ref_default_policy, lock_default_policy>	ownership_default_policy;
typedef no_arithmetic_semantic_policy									arithmetic_default_policy;
#endif // not SMART_PTR_EXCLUDE_DEFAULT_POLICY_SPECIFICATIONS_

/*! 
@brief smart_ptr_base is the base class for smart_ptr.
The main purpose for the smart_ptr_base is to make smart_ptr exception
safe.  When an exception is thrown, only objects that have been constructed
gets their destructor's called.  A base class gets constructed before
the derived class members.  If the base class gets constructed, and 
a the derived class member throws an exception, the base class destructor
is guaranteed to be called.
If no exception is thrown, than the base class destructor does nothing, 
because smart_ptr will nullify the pointer in it's destructor.
This insures that a memory leak does not occur if smart_ptr's members
throw an exception in their constructor.
*/
template<class T, class ALLOCATOR_POLICY>
class smart_ptr_base
{
protected:
	typedef T * ( *clone_fct_Type ) (const T *, bool, void*, ALLOCATOR_POLICY*, void*);//Last two values are not used (See note#1 at end of file)
	T* m_type;
	clone_fct_Type m_clone_fct;
	inline smart_ptr_base(T* type, clone_fct_Type clone_fct): m_type(type), m_clone_fct(clone_fct){}
	inline ~smart_ptr_base() throw()
	{
		if (m_type) m_clone_fct(m_type, false, NULL, NULL, NULL);
	}
private:
	smart_ptr_base(const smart_ptr_base&);
	smart_ptr_base& operator=(const smart_ptr_base& Src);
};


/*! @class smart_ptr
@brief smart_ptr is a smart pointer policy class that can use different 
ownership logic and semantic polices, which allows the developer to
get the best performance and/or interface for a particular requirement.
@section Description
The smart_ptr class can be used with STL containers to create containers 
of smart pointers, moreover it can be used to create a container of 
abstract based objects via smart_ptr.
In general, smart_ptr is faster than boost::shared_ptr.  When used 
with STL containers, the smart pointer is faster than the boost 
pointer containers.  More importantly, the interface for an STL container 
of smart_ptr is a 100% compatible with STL containers, which is not the 
case with the boost pointer containers.
smart_ptr has a policy that allows it to synchronize access to both
the smart pointer and the pointee.
The smart_ptr has been compiled and tested on VC++ 6.0, 7.1, 8.0, GNU 3.x,
Borland 5.5, and Comeau 4.3.
\n
@section  Why_Use_smart_ptr Why use smart_ptr
The short answer is use smart_ptr because it's safer, faster,
more generic, more STL compatible, more adaptable, and more functional.
In general, it's better to use smart pointers over raw pointers, 
because smart pointers automatically handle memory management, and therfore
reduce memory leaks.
Through the policy based designed, smart_ptr can be used for unique 
coding requirements.  However, for most generic requirements,
it can be used without specifying any particular policy, because it has 
default policies for the most common general use case.
Unlike std::auto_ptr, smart_ptr can be used with STL containers.
smart_ptr can out perform other smart pointers like boost::shared_ptr and
boost pointer containers.\n\n
For more information about why you should
prefer using smart_ptr over the boost::shared_ptr and boost pointer 
containers, see following section \ref subsection_note_prefer1
\n
@section  example_usage Example Usage
@include ../examples/example_smart_ptr_usage.cpp
\n
@section  Policy_Classes Policy Classes
The smart_ptr has six main policy class arguments, and some of the policies themselfs have subpolicies.
Over 250 different types of smart pointers can be created using the included policies.
The following is an overview of the main policies.
\n
@subsection subsection1 OWNERSHIP_POLICY
shared_ptr_policy<ref_???_policy>		--	Similar to boost::shared_ptr, but faster than boost::shared_ptr when used with ref_link_policy or ref_intrusive_policy\n
deep_copy_policy						--	Clone ownership logic, that always performance a deep copy to the pointee in the copy constructor\n
copy_on_write_policy<ref_???_policy>	--	Works like boost::shared_ptr until non-constant access.  When a non-constant access is performed, than it clones if it's being referenced\n
@subsubsection subsubsection1 Sub-Policy
ref_link_policy,  ref_count_policy, and ref_intrusive_policy are sub-policies of copy_on_write_policy and shared_ptr_policy\n
ref_link_policy					--	Faster than refence counting logic, but uses more memory\n
ref_count_policy				--	Similar logic used by boost::shared_ptr.  It uses less memory than ref_link_policy, but in (general) slower than ref_link_policy\n
ref_intrusive_policy			--	Faster than ref_link_policy & ref_count_policy, and requires less memory.  However requires implementation for intrusive_ptr_is_ref, intrusive_ptr_add_ref, intrusive_ptr_release\n
Note:	Although ref_intrusive_policy requires the same two functions as does boost::intrusive_ptr, it has an additional requirement for a intrusive_ptr_is_ref function.\n
This additional function is required to make the policy usable by both shared_ptr_policy and copy_on_write_policy\n
For reference-link VS reference-count VS reference-intrusive logic comparison, see following related link:\n
http://www.boost.org/libs/smart_ptr/smarttests.htm \n
For more information on ownership policies, read following link:\n
http://www.awprofessional.com/articles/article.asp?p=31529&seqNum=5&rl=1 \n
http://stlplus.sourceforge.net/stlplus/docs/smart_ptr.html \n

@subsection subsection2 ALLOCATOR_POLICY
The allocator policy class can be use to perform unique allocation/deallocation logic like COM or file-mapping memory.\n
However, here it's also used to link a policy with different types of clone methods.\n
Main allocator policy:\n
allocator_default_policy				--	Uses the type that is passed to the smart_ptr constructor, in order to clone the object\n
clone_function_allocator_policy			--	Uses a virtual do_clone() function in order to clone the object\n
clone_static_function_allocator_policy	--	Uses a static clone() function in order to clone the object\n
\n
<b>The following semantic policies allows the smart pointer to transfer operator logic to the pointee.</b>\n
\n
@subsection subsection3 COMPARISON_SEMANTIC_POLICY:
value_comparsion_semantic_policy		--	Uses value semantics instead of normal pointer semantics.  This allows the smart pointer to be used with std::map and std::set\n
pointer_comparsion_semantic_policy		--	Uses normal pointer semantics, which is used by most smart pointers like auto_ptr and boost::shared_ptr\n
no_comparsion_semantic_policy			--	This will make sure that a compile error is produce when trying to perform a comparison with a smart pointer\n
\n
@subsection subsection4 STREAM_OP_SEMANTIC_POLICY:
value_stream_operator_semantic_policy	--	Allows operator<< and operator>> to be used on the smart pointer, and it will transfer the logic to the pointee\n
no_stream_operator_semantic_policy		--	Will make sure that a compile error is produced when trying to use stream operators with the smart pointer\n
\n
@subsection subsection5 ARITHMETIC_SEMANTIC_POLICY:
value_arithmetic_semantic_policy		--	Uses value semantics instead of normal pointer semantics.  This allows a type like smart_ptr<string> to work with [pString += "Hello World";]\n
no_arithmetic_semantic_policy			--	Will make sure that a compile error is produced when trying to use with arithmetic operators\n
\n
\verbatim
The arithmetic policy will not compile on VC++ 6.0, and it's excluded via #ifdef
\endverbatim 
@subsection SmartPtrLicense License
Copyright (C) 2005-2005 by David Maisonave (Axter)
This software library is free.\n
Permission to use, copy, modify, distribute this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies.  
David Maisonave (Axter) makes no representations
about the suitability of this software for any purpose.
It is provided "as is" without express or implied warranty.
@subsection subsection_note_prefer1 Why prefer using smart_pr over boost::shared_ptr and boost pointer containers
For those who are not familiar with the boost library, the 
Boost library is a well known and highly reputable portable FREE library.
For more information about Boost, see following link: http://boost.org
There is no existing class in the boost library that offers the same 
features available through smart_ptr class.
However, smart_ptr can be compared to the boost::shared_ptr and the 
boost pointer containers, because it can replace their usage.
In general, smart_ptr is faster than boost::shared_ptr, because
by default, smart_ptr uses reference linking instead of reference
counting.
A container of smart_ptr has a better STL compatible interface than 
does the boost pointer containers, and unlike the boost pointer containers, 
a container of smart_ptr will compile with no compile warnings on VC++ 7.1 & 8.0.
Also, unlike boost pointer containers, the smart_ptr can compile on VC++ 6.0.
The smart_ptr class also has better methods to avoid splicing than does
the boost pointer containers, and it allows for more generic development,
by not forcing the target pointee to have cloning functions.
An STL container of smart_ptr is copiable.  The boost pointer containers have
no copy constructors, so they're not copiable vai copy constructor.
The logic required to copy a boost pointer container performs poorly compared
to using the copy constructor of an STL container of smart_ptr.
The boost pointer containers have many incompatibility issues with the
STL containers.  These incompatibility issues do not exist in a 
container of smart_ptr.

To have greater compatibility and efficiency, it's better to use an STL
container of smart_ptr than to use the boost pointer containers.\n
For greater efficiency and more flexibility, it's better to use smart_ptr,
than to use boost::shared_ptr.

@section smart_ptr_Members smart_ptr Members
*/


//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//----------------------------------------------------------------------------------------
//Start of smart_ptr class

template<class T, 
	//Available OWNERSHIP_POLICY policies:				shared_ptr_policy<ref_link_policy>, copy_on_write_policy<ref_link_policy>, deep_copy_policy, shared_ptr_policy<ref_count_policy>, copy_on_write_policy<ref_count_policy>, shared_ptr_policy<ref_intrusive_policy>, copy_on_write_policy<ref_intrusive_policy>
	class OWNERSHIP_POLICY = ownership_default_policy , 
	//Available ALLOCATOR_POLICY policies:				allocator_default_policy, clone_function_allocator_policy, clone_static_function_allocator_policy
	class ALLOCATOR_POLICY =  allocator_default_policy,
	//Available CHECKING_POLICY policies:				no_checking_policy, boost_assert_checking_policy
	class CHECKING_POLICY =  checking_default_policy,
	//Available COMPARISON_SEMANTIC_POLICY policies:	value_comparsion_semantic_policy, pointer_comparsion_semantic_policy, no_comparsion_semantic_policy 
	class COMPARISON_SEMANTIC_POLICY = comparison_default_policy,
	//Available STREAM_OP_SEMANTIC_POLICY policies:		value_stream_operator_semantic_policy, no_stream_operator_semantic_policy  (Not supported on VC++6.0)
	class STREAM_OP_SEMANTIC_POLICY = stream_default_policy,
	//Available ARITHMETIC_SEMANTIC_POLICY policies:	no_arithmetic_semantic_policy, value_arithmetic_semantic_policy  (Not support for non-compliant compilers like VC++6.0 and BCC55)
	class ARITHMETIC_SEMANTIC_POLICY = arithmetic_default_policy>
class smart_ptr : smart_ptr_base<T, ALLOCATOR_POLICY>
{
	enum implement_default_object{eYes, eNo};
	mutable OWNERSHIP_POLICY m_ownership_policy;
public:
	/*! @name Constructors and Destructor */
	//@{
	/*! @brief Normal constructor which takes a valid pointer.
		smart_ptr will only clone the type that is pass to the constructor.
	*/
	template<typename T_obj>
		inline smart_ptr(T_obj* type):smart_ptr_base<T,ALLOCATOR_POLICY>(type, get_alloc_func(type))
	{
		CHECKING_POLICY::check_type_pass_to_constructor(type);
	}
	/*! @brief Copy constructor */
	inline smart_ptr(const smart_ptr& Src):smart_ptr_base<T,ALLOCATOR_POLICY>(Src.m_ownership_policy.constructor_lock_policy(Src.m_type), Src.m_clone_fct), m_ownership_policy(Src.m_ownership_policy, Src.m_type, Src.m_clone_fct, m_type){Src.m_ownership_policy.constructor_unlock_policy(Src.m_type);}
	/*! @brief Default constructor needed for std::map */
	smart_ptr(implement_default_object use_default_obj = eYes):smart_ptr_base<T,ALLOCATOR_POLICY>(NULL, NULL)
	{
		if (use_default_obj == eYes)
		{
			smart_ptr& default_obj = get_or_set_default_object();
			default_obj.lock();
			m_ownership_policy.assign(m_type, m_clone_fct, default_obj.m_type, default_obj.m_clone_fct, default_obj.m_ownership_policy, m_ownership_policy);
			default_obj.unlock();
		}
	}
	/*! @brief  Destructor */
	inline ~smart_ptr() throw()
	{
		m_ownership_policy.destructor_lock_policy(m_type);
		CHECKING_POLICY::before_release(m_type);
		m_ownership_policy.release(m_type, m_clone_fct, m_ownership_policy);
		CHECKING_POLICY::after_release(m_type);
		m_type = NULL;
	}
#if !defined(_MSC_VER) || (_MSC_VER > 1200)
	/*! @brief  Constructor for smart pointer derived type smart_ptr\<DerivedT\> */
	template<class CompatibleDerivedT,	class OWNERSHIP_POLICY_T,	class ALLOCATOR_POLICY_T,	class CHECKING_POLICY_T,	class COMPARISON_SEMANTIC_POLICY_T,	class STREAM_OPERATOR_SEMANTIC_POLICY_T,	class ARITHMETIC_SEMANTIC_POLICY_T>
		smart_ptr(const smart_ptr<CompatibleDerivedT, OWNERSHIP_POLICY_T, ALLOCATOR_POLICY_T, CHECKING_POLICY_T, COMPARISON_SEMANTIC_POLICY_T, STREAM_OPERATOR_SEMANTIC_POLICY_T, ARITHMETIC_SEMANTIC_POLICY_T>& Src):smart_ptr_base<T,ALLOCATOR_POLICY>(NULL, NULL)
	{//No clean way to do this, so just clone it
		Src.lock();
		Src.make_clone(m_type, m_clone_fct);
		Src.unlock();
	}
	//@}
	/*! @name Operators */
	//@{
	template<class CompatibleDerivedT,	class OWNERSHIP_POLICY_T,	class ALLOCATOR_POLICY_T,	class CHECKING_POLICY_T,	class COMPARISON_SEMANTIC_POLICY_T, class STREAM_OPERATOR_SEMANTIC_POLICY_T,	class ARITHMETIC_SEMANTIC_POLICY_T>
		smart_ptr& operator=(const smart_ptr<CompatibleDerivedT, OWNERSHIP_POLICY_T, ALLOCATOR_POLICY_T, CHECKING_POLICY_T, COMPARISON_SEMANTIC_POLICY_T, STREAM_OPERATOR_SEMANTIC_POLICY_T, ARITHMETIC_SEMANTIC_POLICY_T>& Src)
	{//No clean way to do this, so just clone it
		if (m_type != Src.c_ptr())
		{
			m_ownership_policy.assignment_lock_policy(m_type);
			CHECKING_POLICY::before_release(m_type);
			m_ownership_policy.release(m_type, m_clone_fct, m_ownership_policy);
			CHECKING_POLICY::after_release(m_type);
			Src.lock();
			Src.make_clone(m_type, m_clone_fct);
			Src.unlock();
		}
		return *this;
	}
#else
	//For VC++6.0 the smart_ptr is only interchangeable through assignment operator
	template<class CompatiblePointerTypeForVC60>
	smart_ptr& operator=(const CompatiblePointerTypeForVC60& Src)
	{//No clean way to do this, so just clone it
		m_ownership_policy.assignment_lock_policy(m_type);
		CHECKING_POLICY::before_release(m_type);
		m_ownership_policy.release(m_type, m_clone_fct, m_ownership_policy);
		if (m_type) m_ownership_policy.assignment_unlock_policy(m_type);
		CHECKING_POLICY::after_release(m_type);
		Src.lock();
		Src.make_clone(m_type, m_clone_fct);
		Src.unlock();
		return *this;
	}
#endif //_MSC_VER noteq 1200
	/*! @brief Assignment operator */
	smart_ptr& operator=(smart_ptr& Src)
	{
		if (m_type != Src.m_type)
		{
			Src.lock();
			m_ownership_policy.assign(m_type, m_clone_fct, Src.m_type, Src.m_clone_fct, Src.m_ownership_policy, m_ownership_policy);
			Src.unlock();
		}
		return *this;
	}
	bool operator! () const{return c_ptr() == 0;}
#ifdef SMART_PTR_SUPPORT_RETURN_TYPE_THROUGH_POLICY_
	// @cond INCLUDE_ALL_OBJS_
	//Return type for arithmetic operators
	typedef typename ARITHMETIC_SEMANTIC_POLICY::template return_type<smart_ptr>::type arith_retrn_type;
	// @endcond 
	//The arithmetic semantic policy might not make it through boost peer review, so for now there's only
	//a limited set for demonstration purposes.
	arith_retrn_type& operator+=(const smart_ptr& Src){return ARITHMETIC_SEMANTIC_POLICY::plus_equal_smart_ptr(*this, Src);}
	template<class T2>	arith_retrn_type& operator+=(const T2& Src){return ARITHMETIC_SEMANTIC_POLICY::plus_equal(*this, Src);}
	arith_retrn_type& operator-=(const smart_ptr& Src){return ARITHMETIC_SEMANTIC_POLICY::minus_equal_smart_ptr(*this, Src);}
	template<class T2>	arith_retrn_type& operator-=(const T2& Src){return ARITHMETIC_SEMANTIC_POLICY::minus_equal(*this, Src);}
	friend arith_retrn_type operator+(const smart_ptr& a, const smart_ptr& b){return ARITHMETIC_SEMANTIC_POLICY::plus(a, b);}
	friend arith_retrn_type operator-(const smart_ptr& a, const smart_ptr& b){return ARITHMETIC_SEMANTIC_POLICY::minus(a, b);}
	//
	// @cond INCLUDE_ALL_OBJS_
	typedef typename OWNERSHIP_POLICY::template return_type<T>::type_pointee ownership_retrn_type_pointee;
	typedef typename OWNERSHIP_POLICY::template return_type<T>::type_ref ownership_retrn_type_ref;
	// @endcond 
	//Operators that will force exclusive pointer
	inline ownership_retrn_type_pointee operator->() 
	{
		m_ownership_policy.pointee_lock_policy(m_type);
		CHECKING_POLICY::on_dereference(m_type);
		return m_ownership_policy.get_non_constant_ptr(m_type, m_clone_fct);
	}
	inline ownership_retrn_type_ref operator*() 
	{
		m_ownership_policy.pointee_lock_policy(m_type);
		CHECKING_POLICY::on_dereference(m_type);
		return *m_ownership_policy.get_non_constant_ptr(m_type, m_clone_fct);
	}
	inline ownership_retrn_type_pointee operator->()const {
		m_ownership_policy.const_pointee_lock_policy(m_type);
		CHECKING_POLICY::on_const_dereference(m_type);
		return m_type;
	}
	inline ownership_retrn_type_ref operator*() const {
		m_ownership_policy.const_pointee_lock_policy(m_type);
		CHECKING_POLICY::on_const_dereference(m_type);
		return *m_type;
	}
#else
	//Operators that will force exclusive pointer
	inline T* operator->() 
	{
		m_ownership_policy.pointee_lock_policy(m_type);
		CHECKING_POLICY::on_dereference(m_type);
		return m_ownership_policy.get_non_constant_ptr(m_type, m_clone_fct);
	}
	inline T& operator*() 
	{
		m_ownership_policy.pointee_lock_policy(m_type);
		CHECKING_POLICY::on_dereference(m_type);
		return *m_ownership_policy.get_non_constant_ptr(m_type, m_clone_fct);
	}
	inline const T* operator->()const {
		m_ownership_policy.const_pointee_lock_policy(m_type);
		CHECKING_POLICY::on_const_dereference(m_type);
		return m_type;
	}
	inline const T& operator*() const {
		m_ownership_policy.const_pointee_lock_policy(m_type);
		CHECKING_POLICY::on_const_dereference(m_type);
		return *m_type;
	}
#endif //SMART_PTR_SUPPORT_RETURN_TYPE_THROUGH_POLICY_
	//@}

	/*! @name Synchronize Methods */
	//@{
	/*! @brief lock function is used with lock policy, and can be called
		manually (directly), or automatically through scope_lock.
	*/
	void lock() const {if (m_type) m_ownership_policy.lock(m_type);}
	/*! @brief unlock function is used with lock policy, and can be called
		manually (directly), or automatically through scope_lock.
	*/
	void unlock() const {if (m_type) m_ownership_policy.unlock(m_type);}
	/*! @brief trylock is an optional synchronization function that can be
		used with the lock policy.  Use islock function to test trylock success.
		Some implementations do not support trylock and islock logic.
	*/
	void trylock() const {if (m_type) m_ownership_policy.trylock(m_type);}
	/*! @brief islock is an optional synchronization function that can be
		used with the lock policy.  The islock function should never be used
		with lock() method, and should only be used with trylock().  Some
		implementations do not support trylock and islock logic.
	*/
	void islock() const {if (m_type) m_ownership_policy.islock(m_type);}// Should only be used with trylock
	//@}


	/*! @name Misc Methods */
	//@{
	/*! @brief For added safety, call set_default_object to set default object before
		using this class as the second type in a std::map container.
	*/
	static void set_default_object(const smart_ptr& NewValue){get_or_set_default_object(&NewValue);}
	/*! @brief Use equal method to test if two pointee's are equal to each other.
		This does not test if the address of the pointers are equal to each other,
		and instead test if what the pointers are pointing to, are equal to each other.
	*/
	template<class T2>	smart_ptr& equal(const T2& Src){
		(*get_ptr()) = (Src);
		return *this;
	}
	/*! @brief The get_function_ptr method is mainly for internal usage, but it can not 
		be made private/protected because policy methods require access to this method.
	*/
	clone_fct_Type get_function_ptr()const{return m_clone_fct;}
	/*! @brief The swap method is not used by any policy, and it's implemented here
		to make this class compatible with generic coding.
	*/
	void swap(smart_ptr<T> & other)throw(){std::swap(m_ownership_policy, other.m_ownership_policy);std::swap(m_type, other.m_type);std::swap(m_clone_fct, other.m_clone_fct);}
	/*! @brief The make_clone method is used to allow other smart pointers access
		to the cloning logic.  One of the main reasons it's required, is because a
		derived T type, would not have access to the private members (m_clone_fct & m_type),
		so without make_clone, it would be difficult to clone smart_ptr's if the T type where not the same.
	*/
	template<class PT, class FPT>	void make_clone(PT*& ptr, FPT& func_ptr) const
	{
		if (m_type && m_clone_fct)
		{
			ptr = m_clone_fct(m_type, true, NULL, NULL, NULL);
			func_ptr = (FPT)m_clone_fct;
		}
		else
			ptr = NULL;
	}
	//@}


	/*! @name Common Pointer Access Methods */
	//@{
	inline T* get_ptr(){return m_ownership_policy.get_non_constant_ptr(m_type, m_clone_fct);}
	inline const T* c_ptr()const{return  m_type;}
	inline const T& c_ref()const{return  *m_type;}
	//@}

	// @cond INCLUDE_ALL_OBJS_
	typedef COMPARISON_SEMANTIC_POLICY comparison_semantic_policy_type;
	typedef T* pointer;
	typedef T& reference;
	// @endcond 
private:
#if !defined(_MSC_VER) || (_MSC_VER > 1200)
/*
Note#1:
The constructor_and_destructor_allocator_function function has three extra arguments that are not used.  These arguments are not required at all in more compliant compilers
like VC++ 7.x and GNU 3.x.  However, for none compliant (pre-standard) compilers like VC++ 6.0 and BCC55, the extra argument declaration is required
so as to be able to fully qualify the function template type.  The argument declarations are needed for these compilers, but the actual argument variables
are not needed.  Anything pass to these last three arguments is discarded.
*/
	template < typename TT, typename DT, typename AL> static TT * constructor_and_destructor_allocator_function( TT *  ptr, bool bConstruct, DT*, AL* , void*) 
	{
		if (!ptr) return NULL;
		if (bConstruct){return AL::allocate(static_cast<const DT*>(ptr));}
		AL::deallocate(ptr);
		return NULL;
	} 
	template<typename T_obj>
		inline static	clone_fct_Type get_alloc_func(T_obj*)
	{
		T * ( *tmp ) (T *, bool, T_obj*, ALLOCATOR_POLICY*, void*) = constructor_and_destructor_allocator_function<T,T_obj,ALLOCATOR_POLICY>;
		return (clone_fct_Type)tmp;
	}
#else
	template<typename T_obj>
		inline static	clone_fct_Type get_alloc_func(T_obj*)
	{
		T * ( *tmp ) (T *, bool, T_obj*, ALLOCATOR_POLICY*, void*) = constructor_and_destructor_allocator_function<T,T_obj,ALLOCATOR_POLICY>;
		if (!tmp) constructor_and_destructor_allocator_function((T*)NULL, false, (T_obj*)NULL, (ALLOCATOR_POLICY*)NULL, NULL);
		return (clone_fct_Type)tmp;
	}
#endif //_MSC_VER noteq 1200

	static smart_ptr& get_or_set_default_object(const smart_ptr* NewValue = NULL)
	{
		static smart_ptr DefaultObj(eNo);
		if (NewValue && NewValue->m_type)
			DefaultObj = *NewValue;
		return DefaultObj;
	}

#if !defined(_MSC_VER) || (_MSC_VER > 1200)
	template<class T1> friend bool operator==(const smart_ptr & a, const T1 & b){return T1::comparison_semantic_policy_type::is_equal(a,b);}
	template<class T1> friend bool operator!=(const smart_ptr & a, const T1 & b){return !T1::comparison_semantic_policy_type::is_equal(a,b);}
	template<class T1> friend bool operator> (const smart_ptr & a, const T1 & b){return T1::comparison_semantic_policy_type::greater_than(a,b);}
	template<class T1> friend bool operator< (const smart_ptr & a, const T1 & b){return T1::comparison_semantic_policy_type::less_than(a,b);}
	template<class T1> friend bool operator>=(const smart_ptr & a, const T1 & b){return T1::comparison_semantic_policy_type::greater_than_or_equal(a,b);}
	template<class T1> friend bool operator<=(const smart_ptr & a, const T1 & b){return T1::comparison_semantic_policy_type::less_than_or_equal(a,b);}
#else //Can not support exchangeable comparison in VC++ 6.0 
	friend inline bool operator==(const smart_ptr & a, const smart_ptr & b){return COMPARISON_SEMANTIC_POLICY::is_equal(a,b);}
	friend inline bool operator!=(const smart_ptr & a, const smart_ptr & b){return !COMPARISON_SEMANTIC_POLICY::is_equal(a,b);}
	friend inline bool operator< (const smart_ptr & a, const smart_ptr & b){return COMPARISON_SEMANTIC_POLICY::less_than(a,b);}
	friend inline bool operator> (const smart_ptr & a, const smart_ptr & b){return COMPARISON_SEMANTIC_POLICY::greater_than(a,b);}
	friend inline bool operator<=(const smart_ptr & a, const smart_ptr & b){return COMPARISON_SEMANTIC_POLICY::less_than_or_equal(a,b);}
	friend inline bool operator>=(const smart_ptr & a, const smart_ptr & b){return COMPARISON_SEMANTIC_POLICY::greater_than_or_equal(a,b);}
#endif //not defined(_MSC_VER) || (_MSC_VER > 1200)

#if __GNUC__ == 2 && __GNUC_MINOR__ <= 96
// Resolve the ambiguity between our op!= and the one in rel_ops
	template<class T> friend bool operator!=(const smart_ptr<T> & a, const smart_ptr<T> & b){return !COMPARISON_SEMANTIC_POLICY::is_equal(a,b);}
#endif

	template<class T1> T1& input_stream(T1 &is){return STREAM_OP_SEMANTIC_POLICY::input_stream(is, *this);}
	template<class T1> T1& output_stream(T1 &io) const {return STREAM_OP_SEMANTIC_POLICY::output_stream(io, *this);}

	template<class TS> inline friend TS& operator >>(TS &is,smart_ptr &obj){return obj.input_stream(is);}
	template<class TS> inline friend TS& operator <<(TS &io,const smart_ptr &obj){return obj.output_stream(io);}

};

// @cond INCLUDE_ALL_OBJS_

//The following smart_ptr_type, allows typedef declarations for certain types
//Example:
//typedef smart_ptr_type<copy_on_write_policy<ref_link_policy, intrusive_lock_policy> > MyCowSyncSmartPtr;
//MyCowSyncSmartPtr::to<Shape>::type pShape1;
template< 
//Available OWNERSHIP_POLICY policies:				shared_ptr_policy<ref_link_policy>, copy_on_write_policy<ref_link_policy>, deep_copy_policy, shared_ptr_policy<ref_count_policy>, copy_on_write_policy<ref_count_policy>, shared_ptr_policy<ref_intrusive_policy>, copy_on_write_policy<ref_intrusive_policy>
class OWNERSHIP_POLICY = ownership_default_policy , 
	//Available ALLOCATOR_POLICY policies:				allocator_default_policy, clone_function_allocator_policy, clone_static_function_allocator_policy
class ALLOCATOR_POLICY =  allocator_default_policy,
	//Available CHECKING_POLICY policies:				no_checking_policy, boost_assert_checking_policy
class CHECKING_POLICY =  checking_default_policy,
	//Available COMPARISON_SEMANTIC_POLICY policies:	value_comparsion_semantic_policy, pointer_comparsion_semantic_policy, no_comparsion_semantic_policy 
class COMPARISON_SEMANTIC_POLICY = comparison_default_policy,
	//Available STREAM_OP_SEMANTIC_POLICY policies:		value_stream_operator_semantic_policy, no_stream_operator_semantic_policy  (Not supported on VC++6.0)
class STREAM_OP_SEMANTIC_POLICY = stream_default_policy,
	//Available ARITHMETIC_SEMANTIC_POLICY policies:	no_arithmetic_semantic_policy, value_arithmetic_semantic_policy  (Not support for non-compliant compilers like VC++6.0 and BCC55)
class ARITHMETIC_SEMANTIC_POLICY = arithmetic_default_policy>
struct smart_ptr_type
{//Future version of this class will have struct 'to' removed, and ptr_to, might get renamed to 'to'
	template<class T> struct to{typedef smart_ptr<T, OWNERSHIP_POLICY, ALLOCATOR_POLICY, CHECKING_POLICY, COMPARISON_SEMANTIC_POLICY, STREAM_OP_SEMANTIC_POLICY, ARITHMETIC_SEMANTIC_POLICY> type;};
	template<class T> struct ptr_to : public smart_ptr<T, OWNERSHIP_POLICY, ALLOCATOR_POLICY, CHECKING_POLICY, COMPARISON_SEMANTIC_POLICY, STREAM_OP_SEMANTIC_POLICY, ARITHMETIC_SEMANTIC_POLICY>{};
	template<class T1> inline smart_ptr_type(T1 &t1): smart_ptr(t1){}
};


struct smart_ptr_policies  //Common policies
{
	typedef smart_ptr_type<copy_on_write_policy<ref_link_policy, intrusive_lock_policy> >	cow_sync_ptr; 
	typedef smart_ptr_type<deep_copy_policy<intrusive_lock_policy> >						copy_sync_ptr; 
	typedef smart_ptr_type<shared_ptr_policy<ref_link_policy, intrusive_lock_policy> >		shared_sync_ptr; 
	typedef smart_ptr_type<copy_on_write_policy<ref_link_policy> >							cow_ptr; 
	typedef smart_ptr_type<deep_copy_policy<> >												copy_ptr; 
	typedef smart_ptr_type<shared_ptr_policy<ref_link_policy> >								shared_ptr; 
};
// @endcond 

/*! @page  Revision_History Revision History
@section VersionToRevisionNumberSwitch Version to Revision Number Switch
Revision History did not start until after version .21.
Previous versions of smart_ptr were incremented by version numbers.
After the code was added to source control, the version numbers
were abandon, and revision numbers are now used.
@section RevisionHistorySection History
\verbatim
$Date: 3/23/06 3:48a $	(Last Update)
$Revision: 8 $			(Last Revision)
$Log : $
$History: smart_ptr.hpp $
 * 
 * *****************  Version 8  *****************
 * User: Axter        Date: 3/23/06    Time: 3:48a
 * Updated in $/SmartPointers/Smartptr
 * Changed assign logic for ref link and ref instrusive, because of
 * previous change in COW
 * 
 * *****************  Version 7  *****************
 * User: Axter        Date: 3/22/06    Time: 2:07a
 * Updated in $/SmartPointers/Smartptr
 * Fixed bug with ref count and COW policy
 * 
 * *****************  Version 6  *****************
 * User: Axter        Date: 3/06/06    Time: 12:50a
 * Updated in $/SmartPointers/Smartptr
 * 
 * *****************  Version 5  *****************
 * User: Axter        Date: 3/06/06    Time: 12:13a
 * Updated in $/SmartPointers/Smartptr
 * 
 * *****************  Version 4  *****************
 * User: Axter        Date: 3/05/06    Time: 11:04a
 * Updated in $/SmartPointers/Smartptr
 * Updated comments
 * 
 * *****************  Version 3  *****************
 * User: Axter        Date: 3/01/06    Time: 7:12p
 * Updated in $/SmartPointers/smartptr
 * 
 * *****************  Version 2  *****************
 * User: Axter        Date: 2/27/06    Time: 10:12p
 * Updated in $/SmartPointers/smartptr
 * Removed explicit keyword from copy constructor of derived type.  If the
 * explicit keyword is used, the smart_ptr is not able to take an input
 * smart_ptr argument using a different policy type.
 * 
 * *****************  Version 1  *****************
 * User: Administrator Date: 2/27/06    Time: 7:04a
 * Created in $/SmartPointers/smartptr

\endverbatim 
*/
#endif //not smart_ptr_HPP_HEADER_GUARD_

